/* ====================================================================================
HEALPix-related local routines.
==================================================================================== */
/*
			Fetch header files.
*/
#include <iostream>
#include <math.h>
#include "heal.h"
#include "map_exception.h"

using namespace std;
/*
			Constants.
*/
const double lambda0 = 0;	// offset
/* ----------------------------------------------------------------------------
'xy2pix' converts an x-y coordinate into a pixel number.

Arguments:
	ix,iy - The x-y coordinates.

Returned:
	pix - The pixel number.

Written by Nicholas Phillips.
---------------------------------------------------------------------------- */
long xy2pix(long ix, long iy)
{
	static int x2pix[128];
	static int y2pix[128];
	static bool xy2pix_called = false;
	if( ! xy2pix_called ) {
		mk_xy2pix(x2pix, y2pix);
	xy2pix_called = true;
	}
	//int ix_low = (int)fmod(ix,128);
	int ix_low = ix % 128;
	int ix_hi  = ix / 128;
	int iy_low = iy % 128;
	int iy_hi  = iy / 128;
	return (x2pix[ix_hi]+y2pix[iy_hi]) * (128 * 128)
			+ (x2pix[ix_low]+y2pix[iy_low]);
}
/* ----------------------------------------------------------------------------
'toMollweide' converts phi/lambda into x/y and theta, implementing a conversion
into the Molleweide projection.

Arguments:
	phi,lambda - The input coordinates (colatitude and longitude), in radians.
	x,y        - The output cartesian coordinates.

Returned:
	theta - The auxiliary angle.

Written by Nicholas Phillips.
---------------------------------------------------------------------------- */
double toMollweide(const double phi, const double lambda, double &x, double &y)
{
	const double r2 = sqrt(2.);
	double theta;
	x = phi;
	y = lambda;
/*
			Start by solving 2theta + sin(2theta) == pi sin(phi), for theta

				Upper limit.
*/
	if( fabs(x - M_PI/2) < 1e-6) {
		theta = M_PI/2;
	}
/*
				Lower limit
*/
	else if (fabs(x + M_PI/2) < 1e-6 ) {
		theta = -M_PI/2;
	}
	else {
/*
				Iterative computation using Newton's method.
*/
		const int nmax = 20;		// max number of iterations for Newton's method
		const double eps = 1e-6; 	// how well we solve for theta
		int i;
		double dtheta;
		double f,df;
		theta = x/2;									// The initial guess.
		f = 2*theta + sin(2*theta) - M_PI*sin(phi); 	// Solve for this function.
		for(i = 0; i < nmax; i++) {
			if( fabs(f) < eps)
				break;
			df = 2*(1+cos(2*theta));
			dtheta = -f/df;
			theta += dtheta;
			f = 2*theta + sin(2*theta) - M_PI*sin(x);
		}
	}
/*
			Now that we have the theta, we can compute the cartesian coordinates.
*/
	x = 2. * r2 * (y - lambda0) * cos(theta) / M_PI;
	y = r2 * sin(theta);
	
	return theta;
}
/* ----------------------------------------------------------------------------
'fromMollweide' converts x/y into phi/lambda, implementing a conversion from
the Molleweide projection.

Arguments:
	x,y        - The input cartesian coordinates.
	phi,lambda - The output coordinates (colatitude and longitude), in radians.

Returned:
	theta - The auxiliary angle.

Written by Michael R. Greason, ADNET, 05 September 2007.
---------------------------------------------------------------------------- */
double fromMollweide(const double x, const double y, double &phi, double &lambda)
{
	const double r2 = sqrt(2.);
	const double pi2 = M_PI / 2.;
	double theta;
	if (fabs(y) > r2) throw MapException(MapException::Other);
	theta  = asin(y / r2);
	phi    = pi2 - asin(((2. * theta) + sin(2. * theta)) / M_PI);
	lambda = lambda0 + ((M_PI * x) / (2. * r2 * cos(theta)));
	return theta;
}
